{{
********************************************
* jvmVpCore - Virtual Peripheral functions *
********************************************
}}


{{

PASM Newbie alert! Snapshot alert!

This is not a request for coding critique. Suggestions for real improvements
rather than addressing subjective cosmetic cruft would be appreciated.


Hi, If you're reading this, you've downloaded a work that may or may not be
included in the Propeller JVM. I have had some experiences with assembly over
my 20 year software engineering career; they were mostly in "support of" large
embedded systems and multiprocessor designs to get C/C++ booted, for gdb, or
for ISRs handling.

My successful career leads me to believe I can cross any river or ocean, and
thus became involved with this after toying with propeller for some projects.

This is my first PASM attempt. It is not optimized, nor is it complete.
The TIMER and PWM work; the ADC, DAC, and UART devices are to be implemented.
I find it unlikely that all of this can happen in just one COG in 8.68us 8-)
Thus, this framework if possible may need replication to other COGS or be
done in a different way. All timings are verified on an O'scope.

This code is not currently added to JVM build. I am encouraged to make it
public since it contains requirements documentation and a general framework
that may be included in the JVM. I have tested a variant of Timer with the
JVM. I will add this to JVM later pending some issue resolution.

The next part of the task I plan to execute is the DAC since it is very
similar to the PWM.

--Steve, March 2, 2008 ...
 
}}

{{
/**
 *
 * Implementing Virtual Peripherals (VP) requires a process to run in the background.
 * This is a good opportunity for Propeller, although it takes another COG out of
 * the user's hands. We don't have a strategy for the user to control a cog yet.
 * One possibility is extending the VP set to more cogs. Another is to provide a
 * way for a user binary to execute on a cog. These are TBD of course.
 *
 * The elements that need a background process are: ADC, DAC, PWM, Timer, and UART.
 * It is expected that one Propeller COG can behave as the VP thread which does
 * processing ever 8.68us. The goal is to have all VP's be processed in that time*.
 *
 * *Note: It is very likely that more than one COG should be used to realize the full
 * potential of the VP design. Having cog/hub shared memory access is very expensive
 * and may not allow time for even the best asm programmer to do everything in 8.68us.
 * Using a time-multiplex design would make all features possible, but this would reduce
 * the granularity of events.
 *
 * One can argue that certain functions such as the ADC do not need a timer process,
 * but the conversion can be made more accurate if done in time slices.  The DAC
 * must be done as part of a timer process since it is more or less just a PWM.
 *
 * Common VP design requirements:
 *
 * - Up to 6 VP should run at once.
 * - All VP elements should run every 8.68us.
 * - VP pin usage is mutually exclusive.
 *
 * - Requirement is to have up to 6 PWM or 6 ADC or 6 DAC or 6 Uart run at a time.
 *   The java code clearly uses only 1 timer at any given moment, the VP of the
 *   timer can be assigned and would replace one of the other VPs in the list above.
 *
 * - jvmNative method nm_core_CPU_uninstall stops a VP this is a good place to
 *   arrange notification for the cog to not update elements.
 *
 * Design requirements for ADC:
 *
 * - ADC will be initialized by native call.
 * - ADC value should be readable by user. Variable for this written by COG/read by java.
 * - ADC pins should be selectable one for input one for reference output.
 * - External RC network is required on the input pin for ADC to function.
 * - Input measurements are limited from 0 to 3.3VDC.
 * - Up to 6 ADC can be used at once; the code will update measurments on each 8.68us tick.
 *
 * Design requirements for DAC:
 *
 * - DAC will be initialized by native call.
 * - DAC should be writeable by user. Variable for this written by java, read by COG.
 * - DAC output pins should be selectable.
 * - External RC network is required on the output pin for DDC to function.
 * - Output measurements are limited from 0 to 3.3VDC.
 * - Up to 6 DAC can be used at once; the code will update output on each 8.68us tick.
 *
 * Design requirements for PWM:
 *
 * - PWM pins should be selectable.
 * - PWM will be initialized by native call.
 * - Each PWM pin's pulse cycle characteristics can be updated any time.
 * - Up to 6 PWM can be used at once; the code will update output on each 8.68us tick.
 *
 * Design requirements for Timer:
 *
 * - Timer should tick every 8.68us.
 * - Timer should start when initialzied by native call.
 * - Timer value should latch on demand from native call.
 * - The Propeller COG process main loop with wait is the basis for the timer.
 * - Timer variables will be incremented by COG and read by java when running.
 * - Timer latch message can be sent from java to COG and cleared by COG for java read.
 *
 * Design requirements for Uart:
 *
 * - Uart will operate in one direction per instance.
 * - Uart will buffer input or output at 256 bytes each.
 * - Uart receiver using HW flow control will stall sender when <= 16 RX buffer bytes are free.
 * - Uart receiver using HW flow control will remove back-pressure when > 16 bytes are free.
 * - Uart baud rate will be selectable from 600 to 57600 BPS.
 * - Uart stop bits will be selectable from 1 to 8.
 * - Uart level sense can be inverted for both data bits and control lines.
 * 
 * - Up to 6 uni-directional Uarts will be allowed.
 * 
 */

/**
 *
 * COG Initialization and Start/Stop design
 *
 * The jvmVpCore start method should be called at system startup (jd.init would be good).
 * Once started, the cog should run freely although stop is possible since all
 * initialization required for cog function should be put into start.
 *
 * VP Start/Stop. VPs are started by the given init or install method below.
 * Calls to jvmVpCore vp.startVP(addr, bank, type) can be added to these methods.
 *
 * nm_core_Uart_txInit(bank)
 * nm_core_Uart_rxInit(bank)
 * nm_core_Timer_init(bank)
 * nm_core_PWM_install(bank)
 * nm_core_DAC_install(bank)
 * nm_core_ADC_install(bank)
 *
 * All VPs are stopped individually by the nm_core_CPU_uninstall method.
 * Call to jvmVpCore vp.stopVP(bank) has been added.
 *
 */
 
/**
 * Timer design -- implemented
 *
 * This is the easiest design as it involves "generally" incrementing the counter
 * on every pass of the waitcnt timeout in the main cog loop. More detail follows.
 *
 * Timer should start when initialzied by native call. At start time the nmTimer
 * registers will be cleared by java code. This init call is our opportunity to
 * know that the timer is reset, and reset our internal timer accordingly.
 *
 * The nmTimer register is written by the JVM/Java before startup. Afterwards
 * only the COG should write the register and JVM/Java read only the register
 * until the interface is disabled. When disabled, the COG will not access any register.
 * The nmLatchTimer register is written by cog, and read only by JVM/Java.
 * The nmLatchItMsg is set by JVM/Java to signal a timer latch; once written this
 * this value should not be written again until after it is read as clear by JVM/Java. 
 *
 * To prevent a deadlocks the COG and JVM/Java should obey the access rules.
 *
 * The cog will clear its internal timer on init and update latches on each tick.
 *
 * ' file stamp\core\Timer.java (offsets within a VP bank) 
 * nmPeriphISRVec = $00  '// offset for all VPs indicates the bank number VPTYPE*0x10
 *
 * This nmLatchTimer register set is not touched by cog according to jvmNative use.
 *
 * nmLatchTimer1  = $08  '// LSB of 32 bit timer latch
 * nmLatchTimer2  = $09  '
 * nmLatchTimer3  = $0A  '
 * nmLatchTimer4  = $0B  '// MSB of 32 bit timer latch ... not touched by cog
 *
 * This nmTimer register set is written once by Java/JVM at startup and accessed
 * as read or write by the process. Shitty design alert. At least java start
 * writes the counter before starting the timer. This is Little Endian like Prop.
 * 
 * nmTimer1       = $04  '// LSB of 32 bit timer ... read by cog / write by JVM
 * nmTimer2       = $05  '
 * nmTimer3       = $06  '
 * nmTimer4       = $07  '// MSB of 32 bit timer ... read by cog / write by JVM
 *
 * nmLatchItMsg   = $02
 * 
 * '///////// these registers are redefined to allow long access for Timer //////////
 */
 
/**
 * PWM design -- implemented
 *
 * Only the initial state is known to not work at this time; it may be added later.
 *
 * The PWM is implemented according to the following rules:
 *
 * @param highTime the number of 8.68us the waveform should be high.
 * @param lowTime the number of 8.68us the waveform should be low.
 * @param highZero Boolean value (always true) that tells wether highTime != 0.
 * @param lowZero Boolean value (always true) that tells wether lowTime != 0.
 * The zero counter is niether set nor read by the COG.
 *
 * The registers below are read-only by COG and written by JVM/Java:
 *
 * 'file stamp\core\PWM.java (offsets within a VP bank) 
 * nmPWMPort     = $0E  '// 0 for PortA, 1 for PortB (reserved: 2 for PortC, 3 for PortD)
 * nmPWMPin      = $0F  '// bitmask (1 of 8 set)
 *
 * nmPWMhighTime = $0C
 * nmPWMlowTime  = $0D
 *
 * nmPWMhighZero, nmPWMlowZero, count, and state are ignored
 *
 * ///////// these registers are redefined to allow long access for PWM //////////
 *
 */
 
/**
 * ADC Design -- not implemented
 *
 * nmAdcValue    = $0b  '// current ADC value
 * nmInPort      = $0c  '// 0 for port0, 1 for port1
 * nmInPin       = $0d  '// bitmask (1 of 8 set)
 * nmOutPort     = $0e  '// 0 for port0, 1 for port1
 * nmOutPin      = $0f  '// bitmask (1 of 8 set)   
 */
 
/**
 * DAC Design -- not implemented, but would be very similar to PWM
 *
 * nmDacValue    = $0b  '// current ADC value
 * nmDacPort     = $0e  '// 0 for port0, 1 for port1
 * nmDacPin      = $0f  '// bitmask (1 of 8 set)   
 */
 
/**
 * UART Design -- not implemented.
 *
 * 'file stamp\core\Uart.java (offsets within a VP bank) 
 * nmUartDataH    = $02  '// transmit register highbyte
 * nmUartDataL    = $03  '// transmit register lowbyte 
 * nmUartInvert   = $05  '// 0 for normal, 1 for invert
 * nmUartBaud     = $06  '// 1000000/(bps * 4.34)
 * nmUartStopBits = $07  '// Tx only: 00000001 for 1 stop bit, 00000011 for 2 etc
 * nmUartPort     = $0C  '// 0 for PortA, 1 for PortB (reserved: 2 for PortC, 3 for PortD)
 * nmUartPin      = $0D  '// bit mask (1 of 8 set)
 * nmUartHSPort   = $0E  '// 0 for PortA, 1 for PortB (reserved: 2 for PortC, 3 for PortD)
 * nmUartHSPin    = $0F  '// bitmask (1 of 8 set)
 * nmUartBufferH  = $0A  '// address of 256 byte buffer (aligned to $XX00) 
 * nmUartBufferL  = $0B
 * nmUartHead     = $08  '//Index of the next free byte in the buffer. 
 * nmUartTail     = $09  '//Index of the oldest byte in the buffer.
 *                       '//If tailPointer == headPointer the buffer is empty.
 * 'Uart baud constants
 * speed600   = 192  'The speed value for 600 baud.
 * speed1200  = 96   'The speed value for 1200 baud.
 * speed2400  = 48   'The speed value for 2400 baud.
 * speed4800  = 24   'The speed value for 4800 baud.
 * speed7200  = 16   'The speed value for 7200 baud.
 * speed9600  = 12   'The speed value for 9600 baud.
 * speed14400 = 8    'The speed value for 14400 baud.
 * speed19200 = 6    'The speed value for 19200 baud.
 * speed28800 = 4    'The speed value for 28800 baud.
 * speed38400 = 3    'The speed value for 38400 baud.
 * speed57600 = 2    'The speed value for 57600 baud.
 */
  
}}


CON
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000
  '_clkmode = xtal1 + pll8x
  '_xinfreq = 10_000_000

  VP_TICKCLK  = 662  'this value with waitcount makes 8.68us tick period
  VP_TICKMUL  = 10_000_000
  VP_DEBUG_PIN = 1
{{
/**
 * @enum VPTypeEnum
 * VP_NONE=0,  VP_TIMER, VP_ADC, VP_DAC, VP_PWM, VP_UART_RX, VP_UART_TX
 */
 _ typedef enum VPTypeEnum_t
}}
  #0, VP_NONE, VP_TIMER, VP_ADC, VP_DAC, VP_PWM, VP_UART_RX, VP_UART_TX

OBJ
  def: "jvmDefines"
  da:  "jvmVpData"
  'sio: "jvmSerial"
  'dbg: "PASDebug" '<---- Added for Debugger 

PUB main
  'dbg.start(31,30,@entry)
  startVP(da.getVpRamAddr, def#nmPeriph0Bank, VP_TIMER)
  startPWM(da.getVpRamAddr, def#nmPeriph5Bank, 0, 3, 5, 4)
  repeat
  
PUB start(ramaddr)

  if cog <> 0
    return ' dont start more than once

  vpRamPtr := ramaddr 'da.getVpRamAddr
  cog := cognew(@entry, vpRamPtr) + 1
  
    
PUB stop
  if cog
    cogstop(cog~ - 1)
  cog := 0

PUB startPWM(addr, bank, port, pin, hcnt, lcnt)

  start(addr)
  'byte[vpRamPtr + bank - def#nmPeriph0Bank + def#nmPWMState] := 0 'start in low state
  ' initial state not implemented.
  
  byte[vpRamPtr + bank - def#nmPeriph0Bank + def#nmPWMlowTime] := lcnt
  byte[vpRamPtr + bank - def#nmPeriph0Bank + def#nmPWMhighTime] := hcnt
  byte[vpRamPtr + bank - def#nmPeriph0Bank + def#nmPWMPort] := port

  case byte[vpRamPtr + bank - def#nmPeriph0Bank + def#nmPWMPort] 
    1: pin += 8
    2: pin += 16
    3: pin += 24
    
  byte[vpRamPtr + bank - def#nmPeriph0Bank + def#nmPWMPin] := pin 'save propeller pin format
  outa[pin]~  'start low  
  dira[pin]~~ 'make pin output

  startVP(vpRamPtr, bank, VP_PWM)

PUB startVP(addr, bank, type)
{{
/**
 * @fn startVP
 * Call from native init/install method to start VP
 * jvm assembler programmer should make an asm version of this.
 * @param bank - bank number to start
 * @param type - type of VPTypeEnum:
 *   VP_TIMER, VP_ADC, VP_DAC, VP_PWM, VP_UART_RX, VP_UART_TX
 * @note starting with type VP_NONE will not start any VP.
 */
 _ void startVP(int bank, VpTypeEnum type);
}}
  start(addr)
  result := bank
  result >>= 4
  if result => $8 and result =< $D
    result &= $7 ' apparently bank 0 is 8. Num 7 ($70) is defined as "block", but not used
    byte[vpRamPtr + bank - def#nmPeriph0Bank + def#nmPeriphISRVec] := type
{
  sio.str(string($d,"VP Started. Type "))
  sio.hex(type,2)
  sio.str(string(" vpRamPtr "))
  sio.hex(vpRamPtr, 4) 
  sio.str(string(" type addr "))
  sio.hex(vpRamPtr + bank - def#nmPeriph0Bank + def#nmPeriphISRVec, 4) 
  sio.str(string(" type  "))
  sio.hex(byte[vpRamPtr + bank - def#nmPeriph0Bank + def#nmPeriphISRVec], 2) 
}
      
PUB stopVP(bank)
{{
/**
 * @fn stopVP
 * Call nm_core_CPU_uninstall to stop VP.
 * jvm assembler programmer should make an asm version of this.
 * @param bank - bank number to start
 */
 _ void stopVP(int bank);
}}
{{ call nm_core_CPU_uninstall to stop VP }}
  result := bank
  result >>= 4
  if result => $8 and result =< $D
    result &= $7 ' apparently bank 0 is 8. Num 7 ($70) is defined as "block", but not used
    byte[vpRamPtr + bank - def#nmPeriph0Bank + def#nmPeriphISRVec] := false
   
  
DAT
{{
_______________________________________________________________________________
_______________________________________________________________________________

An assembly cog which performs Javelin-like Virtual Peripheral VP functions.

  repeat
    wait 8.68us
    repeat i from 0 to def#MAX_NUM_VPS-1                 'executes every 8.68 usec
}}
                      
entry         org 0

'  --------- Debugger Kernel add this at Entry (Addr 0) ---------
'   long $34FC1202,$6CE81201,$83C120B,$8BC0E0A,$E87C0E03,$8BC0E0A
'   long $EC7C0E05,$A0BC1207,$5C7C0003,$5C7C0003,$7FFC,$7FF8
'  -------------------------------------------------------------- 

              mov     vpRamPtr, par
              'mov     dira, tickout             'set  tickout pin to output
              mov     time, cnt                 'record current time
              add     time, period              'establish next period
              'or      dira, #VP_DEBUG_PIN       'debug only
              'or      dira, #VP_DEBUG_PIN+1     'debug only
              
              'wait until next period ... doesn't matter if wait is here or last
tickLoop      waitcnt time, period
              'xor     outa, tickout             'toggle tick output for frequency measurement
              'xor     outa, #VP_DEBUG_PIN
              'xor     outa, #VP_DEBUG_PIN

              ' since djnz loop expires on 0, do first user request here                            
              mov     vpAsmIndex, #0
              call    #doVpUser 
              'xor     outa, #VP_DEBUG_PIN
              'xor     outa, #VP_DEBUG_PIN

              'do other user requests now
              mov     vpAsmIndex, #5            'use decrement
tickUserLoop  call    #doVpUser
              'xor     outa, #VP_DEBUG_PIN
              'xor     outa, #VP_DEBUG_PIN
              djnz    vpAsmIndex, #tickUserLoop                            

              jmp     #tickLoop                 'loop to next tick

{{
_______________________________________________________________________________
_______________________________________________________________________________
}}

{{
_______________________________________________________________________________
PRI function: doVpUser(vpType)
do VP user stuff here
      vpBank := def#nmPeriph0Bank + (i<<4)               'set currently accessed bank
      type := byte[vpRambank + vpBank + nmPeriphISRVec] 'nmPeriphISRVec holds VP type
      case type
        NONE:   vpNone
        TIMER:  vpTimer
        ADC:    vpAdc
        DAC:    vpDac
        PWM:    vpPwm
        UARTRX: vpUartRx
        UARTTX: vpUartTx

param:  vpType = enumerated type of VP
return: nothing
}}
doVpUser      'first calculate the bank
              mov     vpAsmBank, vpAsmIndex
              shl     vpAsmBank, #4             'shift index << 4
              add     vpAsmBank, vpRamPtr              
              'add     vpAsmBank, #def#nmPeriphISRVec  'nmPeriphISRVec is 0
              'mov     vpAsmType, #VP_TIMER
              rdbyte  vpAsmType, vpAsmBank      'now read from address
              'and     vpAsmType, #$0F           'get first 4 bits only

              'case statement ... make into jump table later
              'if request is zero, we have nothing to do, get out
              tjz     vpAsmType, #doVpUser_ret 
              
              cmp     vpAsmType, #VP_TIMER wz
        if_e  call    #vpAsmTimer

              cmp     vpAsmType, #VP_DAC wz
        if_e  call    #vpAsmAdc

              cmp     vpAsmType, #VP_DAC wz
        if_e  call    #vpAsmDac

              cmp     vpAsmType, #VP_PWM wz
        if_e  call    #vpAsmPwm

              cmp     vpAsmType, #VP_UART_RX wz
        if_e  call    #vpAsmUartRx

              cmp     vpAsmType, #VP_UART_TX wz
        if_e  call    #vpAsmUartTx
         
              'all done now ... get out              
doVpUser_ret  ret


{{
_______________________________________________________________________________
_______________________________________________________________________________
}}


{{
_______________________________________________________________________________
PRI function: vpAsmTimer(...)
The function of the timer is to increment on every "system" tick.
This update function increments once per tick, so increment timer here.
...
PRI vpTimer | t32
  'increment 32bits timer in bank specified by vpBank
...
The 32bit timer register is stored in Little Endian
...
Modifys var1 and var2

}}
vpAsmTimer   'only called when vpType is TIMER
             'note that nmTimer1 is long aligned because vpRambank is long aligned with 2 dummy startbytes
              mov     var1, vpAsmBank           'target bank                
              add     var1, #def#nmTimer1       ' + target offset
              rdlong  var2, var1                ' read timer count          
              add     var2, #1                  'increment timer            
              wrlong  var2, var1                'write new timer count      
vpAsmTimer_ret       ret

{
_______________________________________________________________________________
Keep the word variant around until we're sure the long variant will work
////////////////////////////////////////
This does not include the latching logic.

vpWdAsmTimer   'first check if timer is enabled, if not, don't update
              cmp     vpAsmType, #VP_TIMER wz
    if_ne     jmp     #vpWdAsmTimer_ret          ' if not TIMER service exit method

              'Look for non-zero in nmTimer regs, if zero then timer needs reset
              'since the order doesn't matter, we can use rdword from nmTimer1 offset
              'using rdlong is not possible here unless the address can be on even 4 byte boundary
              mov     var1, vpAsmBank           'target address
              add     var1, #def#nmTimer1       ' + target offset
              rdword  var2, var1                ' read first half  ... only care if zero or not
              tjnz    var2, #vpWdAsmTimer_NoReset
              add     var1, #2                  ' + target offset
              rdword  var2, var1                ' read second half ... only care if zero or not
              tjnz    var2, #vpWdAsmTimer_NoReset

              'if we get here, user wants us to reset timer              
              mov     vpTimerTick, #0

vpWdAsmTimer_NoReset
              'increment timer              
              add     vpTimerTick, #1
              mov     var2, vpTimerTick
              'write new value              
              mov     var1, vpAsmBank           'target address
              add     var1, #def#nmTimer1       ' + target offset
              wrword  var2, var1                'write first half
              add     var1, #2
              shr     var2, #16
              wrword  var2, var1                'write second half
vpWdAsmTimer_ret       ret
}

{{
_______________________________________________________________________________
PRI function: vpAsmAdc(...)
deliver ADC value

}}
vpAsmAdc         '
              mov Parm1, Parm1 
vpAsmAdc_ret     ret


{{
_______________________________________________________________________________
PRI function: vpAsmDac(...)
produce DAC output

}}
vpAsmDac         ' 
              mov Parm1, Parm1 
vpAsmDac_ret     ret


{{
_______________________________________________________________________________
PRI function: vpAsmPwm(...)
produce PWM output

PWM installation is defined by bank ... bank is $80-$DO
    byte[vpRambank + bank + nmPeriphISRVec] := PWM

PWM high/low is controlled by "state" ...
    byte[vpRambank + bank + def#nmPWMHtime + state] ' state = 0 for hi time, 1 for low time

  nmPWMPort     = $0E  '// 0 for PortA, 1 for PortB (reserved: 2 for PortC, 3 for PortD)
  nmPWMPin      = $0F  '// bitmask (1 of 8 set)
  nmPWMhighTime = $0C
  nmPWMlowTime  = $0D

  nmPWMhighZero, nmPWMlowZero, count, and state are ignored

This routine generally follows algoritm from:
  PRI vpPwm | accu,pin,state,count

Modifys varN and:

vpAsmPwmPort  long    0 ' long for debug; for optimization convert to bytes
vpAsmPwmPin   long    0
vpAsmPwmHtime long    0
vpAsmPwmLtime long    0
vpAsmPwmCount long    0
vpAsmPwmState long    0
  
}}
vpAsmPwm      ' see if we need to do work 
              cmp     vpAsmType, #VP_PWM wz
    if_ne     jmp     #vpAsmPwm_ret             'if not PWM service exit method

              'xor     outa, #VP_DEBUG_PIN+1     'debug only
              'xor     outa, #VP_DEBUG_PIN+1     'debug only

              'first, read parameters, then do pwm updates

              ' read port and pin
              mov     var1, vpAsmBank           'target address
              add     var1, #def#nmPWMhighTime  ' hitime is $C
              rdlong  var2, var1                ' save port and pin

              movs    vpAsmPwmHtime, var2       ' save high time
              shr     var2, #8
              movs    vpAsmPwmLtime, var2       ' save low time
              shr     var2, #8
              movs    vpAsmPwmPort, var2        ' save port
              shr     var2, #$8
              movs    vpAsmPwmPin,  var2        ' save pin

              ' set count comparison for hi-time or lo-time
              tjz     vpAsmPwmState, #vpAsmPwmT1
vpAsmPwmT0
              mov     vpAsmPwmComp, vpAsmPwmLtime
              jmp     #vpAsmPwmTdone
vpAsmPwmT1
              mov     vpAsmPwmComp, vpAsmPwmHtime
vpAsmPwmTdone

              ' decrement count ... if non-zero, exit function. else update PWM
              and     vpAsmPwmCount, #$ff
              djnz    vpAsmPwmCount, #vpAsmPwm_Exit

              ' if counter is zero, change pin state
              movs    vpAsmPwmCount, vpAsmPwmComp
              xor     vpAsmPwmState, #1

              ' prepare for output
              mov     var2, #1
              shl     var2, vpAsmPwmPin
              or      dira, var2

vpAsmPwm_St   'if state is zero, do zero part, else do one part
              tjz     vpAsmPwmState, #vpAsmPwm_Do0

              ' do one
              or      outa, var2                ' or to set
              jmp     #vpAsmPwm_Exit
                              
vpAsmPwm_Do0  ' do zero
              xor     var2, var2                ' flip bits, then and to clear
              and     outa, var2

vpAsmPwm_Exit
              'xor     outa, #VP_DEBUG_PIN+1     'debug only
              'xor     outa, #VP_DEBUG_PIN+1     'debug only
                                                    
vpAsmPwm_ret     ret

{{
_______________________________________________________________________________
PRI function: vpAsmUartTx(...)
produce Uart output

}}
vpAsmUartTx      ' 
              mov Parm1, Parm1 
vpAsmUartTx_ret  ret

{{
_______________________________________________________________________________
PRI function: vpAsmUartRx(...)
deliver Uart values

}}
vpAsmUartRx      ' 
              mov Parm1, Parm1 
vpAsmUartRx_ret  ret


{{
_______________________________________________________________________________
_______________________________________________________________________________
}}


{{
_______________________________________________________________________________
DAT ASM Variables
}}

MASK_BYTE     long    $FF
MASK_WORD     long    $FFFF

cog           long    0
tickout       long    |< 25     'APIN=0
infreq        long    5_000_000                 'overloaded by user with start(freq)
period        long    VP_TICKCLK/2 * (VP_TICKMUL / _xinfreq) '115200 Hz (8.68S period) (_clkfreq / period)
time          long    0

vpRamPtr      long    0         'pointer storage set by .spin
vpRamVar      long    0         'not necessary
vpAsmIndex    long    0         'the VP index getting an update
vpAsmBank     long    0
vpAsmType     long    0

var1          long    0
var2          long    0
var3          long    0
var4          long    0
vbit          long    0
Parm1         long    0
Parm2         long    0
RetVar        long    0

vpAsmPwmHtime long    0
vpAsmPwmLtime long    0
vpAsmPwmPort  long    0
vpAsmPwmPin   long    0
vpAsmPwmCount long    0
vpAsmPwmState long    0
vpAsmPwmComp  long    0

{{
_______________________________________________________________________________
_______________________________________________________________________________
}}

              fit     $1F0